function [x,y] = waveStepUpwind(p,x0,u)

    %% Define convenience parameters
    % Drill string
    dxp = p.Lp/p.Pp;
    dtp = dxp/max(p.c_t,p.c_a)*.99 ;  % As per the CFL cond.
    dxc = p.Lc/p.Pc;
    dtc = dxc/max(p.c_t,p.c_a)*.99;  % As per the CFL cond.
    
    % BRI
    dxl = 1/p.Pl;
    dtl = dxl/p.omegaMAX;  % As per the CFL cond.
    
    n = max( ceil(p.dt/min(dtp,dtc)) , ceil(p.dt/dtl) );
    dt  = p.dt/n;
    
    % Total number of drill string cells
    Pt = p.Pp+p.Pc;

    % Parse input
    hp  = x0(1:p.Pp);                % Axial Pipe strain
    hc  = x0(p.Pp+1:Pt);             % Axial Collar strain
    vp  = x0(Pt+1:Pt+p.Pp);          % Pipe Axial velocity
    vc  = x0(Pt+p.Pp+1:2*Pt);        % Collar Axial velocity
    
    fp  = x0(2*Pt+1:2*Pt+p.Pp);      % Angular Pipe strain
    fc  = x0(2*Pt+p.Pp+1:3*Pt);      % Angular Collar strain
    op  = x0(3*Pt+1:3*Pt+p.Pp);      % Pipe angular velocity
    oc  = x0(3*Pt+p.Pp+1:4*Pt);      % Collar angular velocity

    l  = x0(4*Pt+1:4*Pt+p.Pl);       % Depth of cut
    xs = x0(4*Pt+p.Pl+1);            % Shock Sub displacement
    
    % Compute Riemann invariants
    % Axial
    PAalpha = vp+p.c_a*hp; % Pipe
    PAbeta  = vp-p.c_a*hp;
    CAalpha = vc+p.c_a*hc; % Collar
    CAbeta  = vc-p.c_a*hc;
    % Torsional
    PTalpha = op+p.c_t*fp; % Pipe
    PTbeta  = op-p.c_t*fp;
    CTalpha = oc+p.c_t*fc; % Collar
    CTbeta  = oc-p.c_t*fc;

    %%
    t = 0;
    for i=1:n
        %% Topside BC
        if isfield(p,'w0')
            % Constant Hook-load BC
            PAalpha0  = PAbeta(1) + 2*p.w0/p.zeta_ap;
        else
            % Constant axial velocity BC
            PAalpha0  = -PAbeta(1) + 2*p.v0;
        end
        PTalpha0  = -PTbeta(1) + 2*p.omega0;
        
        %% Pipe - collar interface
        Zbar_t = p.zeta_tc/p.zeta_tp;
        PTbetaPp1 = 1/(1+Zbar_t)*( PTalpha(p.Pp)*(1-Zbar_t) + 2*Zbar_t*CTbeta(1) );
        CTalpha0  = 1/(1+Zbar_t)*( 2*PTalpha(p.Pp) - CTbeta(1)*(1-Zbar_t) );

        if isfield(p,'Ks')  % if sub in drill-string
            v_sp = PAalpha(p.Pp)+PAbeta(p.Pp);
            v_sc = CAalpha(1)+CAbeta(1);
            w_s  = -p.Ks*xs - p.Ds*(v_sc-v_sp);

            PAbetaPp1 = PAalpha(p.Pp) - 2*w_s/p.zeta_ap;
            CAalpha0  = CAbeta(1)     + 2*w_s/p.zeta_ac;
        else
            Zbar_a = p.zeta_ac/p.zeta_ap;
            PAbetaPp1 = 1/(1+Zbar_a)*( PAalpha(p.Pp)*(1-Zbar_a) + 2*Zbar_a*CAbeta(1) );
            CAalpha0  = 1/(1+Zbar_a)*( 2*PAalpha(p.Pp) - CAbeta(1)*(1-Zbar_a) );
        end        
        %% Bit rock interaction
        % Bit velocity
        vb = 1/2*(CAbeta(p.Pc)+CAalpha(p.Pc));
%         vb = p.v0;

        ob = 1/2*(CTbeta(p.Pc)+CTalpha(p.Pc));
%         ob = p.omega0;

%         if ob>p.omegaMAX
%             warning('upwind:CFLviolation','ob larger than omegaMAX.');
%         end

        % Normalized depth of cut
        d = max( l(end) , 0);
        
        %% g function nonlinearity
        wc = d*p.Bri_a;   % Axial cutting component
        tc = d*p.Bri_t;   % Torsional cutting component
        
        delta = .01;
        ga = 1/p.Wf * ( p.zeta_ac*CAalpha(p.Pc) - wc - u.wb);
        ga = (vb>delta) + ((vb)<=delta) * min( max(ga,0), 1);

        wf = p.Wf*ga;
        tf = p.Wf*p.beta*ga;

        aStick = (ga<1) && (ga>0);
        
        % Torsional stick
        g_tt = p.zeta_tc/2*(CTalpha(p.Pc)-CTbeta(p.Pc)) - tc - p.Wf*p.beta;
        tStick = (ob<0.02) && (g_tt<0);

        ToB = tc + tf + u.tb; % Total ToB
        WoB = wc + wf + u.wb; % Total WoB

        % BH BC
        if tStick   % Zero velocity BC
            CAbetaPp1    = -CAalpha(p.Pc);
            CTbetaPp1    = -CTalpha(p.Pc);
        elseif aStick
            CAbetaPp1    = -CAalpha(p.Pc);
            CTbetaPp1    =  CTalpha(p.Pc)   - 2*ToB/p.zeta_tc;
        else
            CAbetaPp1    =  CAalpha(p.Pc)   - 2*WoB/p.zeta_ac;
            CTbetaPp1    =  CTalpha(p.Pc)   - 2*ToB/p.zeta_tc;
        end

        % Recompute bit velocity based on update stiction forces
%         vb = 1/2*(Abeta(p.P)+Aalpha(p.P));
%         ob = 1/2*(Tbeta(p.P)+Talpha(p.P));
        
        %% Augment Riemann invariants with interface values
        PAalpha_pad = [PAalpha0; PAalpha];
        PAbeta_pad  = [PAbeta; PAbetaPp1];
        CAalpha_pad = [CAalpha0; CAalpha];
        CAbeta_pad  = [CAbeta; CAbetaPp1];
        PTalpha_pad = [PTalpha0; PTalpha];
        PTbeta_pad  = [PTbeta; PTbetaPp1];
        CTalpha_pad = [CTalpha0; CTalpha];
        CTbeta_pad  = [CTbeta; CTbetaPp1];

        %% 1st order Upwind
        % Drill string
        PAalpha = PAalpha - dt/dxp *p.c_a   *diff(PAalpha_pad) -dt*p.ka*(PAbeta+PAalpha)/2;
        PAbeta  = PAbeta  + dt/dxp *p.c_a   *diff(PAbeta_pad)  -dt*p.ka*(PAbeta+PAalpha)/2;
        CAalpha = CAalpha - dt/dxc *p.c_a   *diff(CAalpha_pad) -dt*p.ka*(CAbeta+CAalpha)/2;
        CAbeta  = CAbeta  + dt/dxc *p.c_a   *diff(CAbeta_pad)  -dt*p.ka*(CAbeta+CAalpha)/2;
        
        PTalpha = PTalpha - dt/dxp *p.c_t   *diff(PTalpha_pad) -dt*p.kt*(PTbeta+PTalpha)/2;
        PTbeta  = PTbeta  + dt/dxp *p.c_t   *diff(PTbeta_pad)  -dt*p.kt*(PTbeta+PTalpha)/2;
        CTalpha = CTalpha - dt/dxc *p.c_t   *diff(CTalpha_pad) -dt*p.kt*(CTbeta+CTalpha)/2;
        CTbeta  = CTbeta  + dt/dxc *p.c_t   *diff(CTbeta_pad)  -dt*p.kt*(CTbeta+CTalpha)/2;

        % Hack to ensure numeric stability of the inclusion
        if aStick || tStick
            CAbeta(p.Pc)   = -CAalpha(p.Pc);
        end
        if tStick
            CTbeta(p.Pc) = - CTalpha(p.Pc);
        end

        % BRI
        % BRI BC is l(0) = 0
        l_pad = [0; l];
        l = l - dt/dxl *max(ob,0)*p.N/(2*pi) *diff(l_pad) + dt*vb;
        
        % Sub ODE
        if isfield(p,'Ks')
            xs = xs + dt*(v_sc-v_sp);
        else
            xs = 0;
        end

        t = t + dt;
    end
    %%

    hp = 1/(2*p.c_a) *(PAalpha-PAbeta);
    hc = 1/(2*p.c_a) *(CAalpha-CAbeta);
    vp = 1/2         *(PAbeta+PAalpha);
    vc = 1/2         *(CAbeta+CAalpha);
    fp = 1/(2*p.c_t) *(PTalpha-PTbeta);
    fc = 1/(2*p.c_t) *(CTalpha-CTbeta);
    op = 1/2         *(PTbeta+PTalpha);
    oc = 1/2         *(CTbeta+CTalpha);
    
    % Parse states
    x(1:p.Pp)                = hp;        % Axial Pipe Strain
    x(p.Pp+1:Pt)             = hc;        % Axial Collar Strain
    x(Pt+1:Pt+p.Pp)          = vp;        % Axial Pipe Velocity
    x(Pt+p.Pp+1:2*Pt)        = vc;        % Axial Collar Velocity
    x(2*Pt+1:2*Pt+p.Pp)      = fp;        % Angular Pipe Strain
    x(2*Pt+p.Pp+1:3*Pt)      = fc;        % Angular Collar Strain
    x(3*Pt+1:3*Pt+p.Pp)      = op;        % Angular Pipe Velocity
    x(3*Pt+p.Pp+1:4*Pt)      = oc;        % Angular Collar Velocity
    x(4*Pt+1:4*Pt+p.Pl)      = l;         % Distributed cut
    x(4*Pt+p.Pl+1)           = xs;        % Shock Sub displacement
    
    % Parse output
    y(1) = d;
    y(2) = 1-ga;
    y(3) = wf;
    y(4) = WoB;
end




















